/*! \file  SEkey.h
 *  \brief This file includes everything about SEkey.
 *  \author Fornero Matteo
 *  \date 22/11/2019
 */

#ifndef SEKEY_H_
#define SEKEY_H_

#include "../sefile/SEfile.h"
#include "../sqlite/sqlite3.h"

#define PINLEN 32 /**< @brief Length (bytes) of the PIN used to login as user or admin to the SEcube. */
#define AES256KEYLEN 32 /**< @brief Length of an AES-256 key expressed in bytes. */
#define IDLEN 11 /**< @brief Maximum length expected (in bytes) for a generic ID (could be a key, a user or a group). The value is 11 because each ID must have 1 literal at the beginning followed by up to 10 numbers. */
#define NAMELEN 100 /**< @brief This is the maximum length accepted for a name or label (i.e. the username, the label of a group or the label of a key). */
#define TRY_LIMIT 5 /**< @brief Maximum number of attempts updating SEkey in user mode. If the limit is reached and the update failed, recovery will be needed. */
#define UPDATE_RECORD_HEADER_LEN 11 /**< @brief Length of header of each update record in the update, init or recovery file. 1 byte for the type, 8 bytes for the counter, 2 bytes for the length. */

/** @brief Record type identifiers for SEkey update files. */
enum update_record_type {
	DELETE_USER_FROM_GROUP, /**< Request to delete a user from a group. */
	DELETE_USER, /**< Request to delete a user from SEkey. */
	KEY_DATA, /**< Request to store a new key inside the SEcube flash memory. */
	SQL_QUERY, /**< Request to execute a SQL query on the database. */
	DELETE_GROUP /**< Request to delete a group from SEkey. */
};

/** @brief Error codes returned by functions of SEkey. */
enum sekey_error {
	SEKEY_OK = 0, /**< Function completed correctly. */
	SEKEY_ERR = 1, /**< Generic error. SEkey database unchanged. */
	SEKEY_FILE_FOUND = 2, /**< File found. */
	SEKEY_FILE_NOT_FOUND = 3, /**< File not found. */
	SEKEY_ERR_AUTH = 4, /**< Requested operation not allowed with current SEkey mode (admin or user). */
	SEKEY_USER_GROUP_DUP = 5, /**< User already belonging to the group. */
	SEKEY_KEY_DUP = 6, /**< Key already stored in SEkey. */
	SEKEY_GROUP_NOT_FOUND = 7, /**< No such group inside SEkey. */
	SEKEY_GROUP_DUP = 8, /**< Group already stored in SEkey. */
	SEKEY_USER_DUP = 9, /**< User already stored in SEkey. */
	SEKEY_USER_NOT_FOUND = 10, /**< No such user inside SEkey. */
	SEKEY_KEY_NOT_FOUND = 11, /**< No such key inside SEkey. */
	SEKEY_ERR_PARAMS = 12, /**< Invalid input parameters. */
	SEKEY_UNCHANGED = 13, /**< Error. SEkey database unchanged. */
	SEKEY_RESTART = 14, /**< Error while executing commit or rollback on the database. Restart the application to go back to coherent database content. */
	SEKEY_CORRUPTED = 15, /**< SEkey database corrupted. */
	SEKEY_REPROG = 16, /**< Error while initializing a SEcube. Device erase needed. */
	SEKEY_RESTART_REPROG = 17, /**< Error while initializing a SEcube. Device erase needed and SEkey host application restart required. */
	SEKEY_BLOCKED = 18, /**< Impossible to update SEkey. SEkey blocked until update is completed. */
	SEKEY_COMMON_GROUP_NOT_FOUND = 19, /**< Returned when the application asks to SEkey to return the key to encrypt data for a set of users but there is no common groups among those users. The application must check for this
	error and, in case this error is returned, proceed to request a key to encrypt data in a unicast manner (1 to 1) instead of multicast (1 to many). */
	SEKEY_UNSUPPORTED = 20, /**< Returned when a certain feature is not yet supported by the system (i.e. requesting a key with a type different from symmetric data encryption). */
	SEKEY_INVALID_KEY = 21, /**< The key is not valid (i.e. does not exist or has invalid attributes). */
	SEKEY_INACTIVE_KEY = 22, /**< The key is not active, so can't be used for encryption. */
	SEKEY_COMPROMISED_KEY = 23, /**< The key is in compromised status. */
	SEKEY_DESTROYED_KEY = 24, /**< The key is in destroyed status. */
	SEKEY_DEACTIVATED_KEY = 25, /**< The key is in deactivated status. */
	SEKEY_PREACTIVE_KEY = 26, /**< The key is in preactive status. */
	SEKEY_SUSPENDED_KEY = 27 /**< The key is in suspended status. */
};

/* WARNING: key ID = 0 is not allowed */
#define NUMBER_RESERVED_KEY_IDS 100 /**< @brief Batch of reserved key identifiers. These can be used without SEkey (i.e. IDs 1 and 2 are used to implement SEkey) and are not managed in any way (key life cycle, etc. not supported). */
#define NUMBER_MASTER_SLAVE_KEY_IDS 100000 /**< @brief Batch of key identifiers reserved to master-slave (admin-user) keys of the SEkey KMS. */
#define KEY_ID_RESERVED_RANGE_BEGIN 1 /**< @brief Minimum value of reserved key identifiers (notice that 1 and 2 are already used by SEkey). */
#define KEY_ID_RESERVED_RANGE_END NUMBER_RESERVED_KEY_IDS /**< @brief Maximum range value of reserved key identifiers (100). */
#define KEY_ID_SEKEY_BEGIN (KEY_ID_RESERVED_RANGE_END + 1)  /**< @brief Minimum range value of standard SEkey key identifiers (101). */
#define KEY_ID_SEKEY_END (UINT32_MAX - NUMBER_MASTER_SLAVE_KEY_IDS)  /**< @brief Maximum range value of standard SEkey key identifiers (2^32 -1 -100000). */
#define KEY_ID_MASTER_SLAVE_BEGIN (KEY_ID_SEKEY_END + 1) /**< @brief Minimum range value of master-slave key identifiers. */
#define KEY_ID_MASTER_SLAVE_END UINT32_MAX /**< @brief Maximum range value of master-slave key identifiers. */
#define SEkey_ID 1 /**< @brief The key ID 1 is reserved for the database internally used by SEkey. */
#define wildcard_key_ID 2 /**< @brief The key ID 2 is reserved for internal purposes of SEkey. */

/** @brief Possible status assumed by a key. There are specific rules for status transition, see check_key_transition_validity() for more informations. */
enum class se_key_status {
	statusmin = 0, /**< Lower boundary. This status is not valid, it is used only to check if the actual status value is within the boundaries. */
	preactive = 1, /**< Default key status (automatically set on key generation). */
    active = 2, /**< The key can be used for encryption and decryption. */
    suspended = 3, /**< The key may be used for decryption, it may also be reactivated in future. */
    deactivated = 4, /**< The key can be used only for decryption. */
    compromised = 5, /**< The key must not be used for encryption. Files encrypted with this key must be decrypted and re-encrypted as soon as possible with a new key. */
    destroyed = 6, /**< The key is not stored anymore in the KMS. Only the metadata are kept. */
	statusmax = 7 /**< Upper boundary. This status is not valid, it is used only to check if the actual status value is within the boundaries. */
};

/** @brief Possible key types. Notice that only symmetric_data_encryption is supported, other types are listed here in case
 *  someone wanted to expand the features of the system. */
enum class se_key_type {
	typemin = 100,
	private_signature = 101,
	public_signature_verification = 102,
	symmetric_authentication = 103,
	private_authentication = 104,
	public_authentication = 105,
	symmetric_data_encryption = 106,
	symmetric_key_wrapping = 107,
	symmetric_RGB = 108,
	symmetric_key_derivation = 109,
	private_key_transport = 110,
	public_key_transport = 111,
	symmetric_key_agreement = 112,
	private_static_key_agreement = 113,
	public_static_key_agreement = 114,
	private_ephemeral_key_agreement = 115,
	public_ephemeral_key_agreement = 116,
	symmetric_authorization = 117,
	private_authorization = 118,
	public_authorization = 119,
	typemax = 120
};

/** @brief Used to identify the different types of files used by the update mechanism of SEkey.
 * The hierarchy is the same, where the init file is the most important update file
 * (the first to be processed), then there is the recovery file and finally the standard
 * update file (used when SEkey is operating normally without errors). */
enum filetype {
	INIT = 1, /**< Used to specify mode = init file in the open_update_file() function. */
	RECOVERY = 2, /**< Used to specify mode = recovery file in the open_update_file() function. */
	NORMAL = 3 /**< Used to specify mode = normal update file in the open_update_file() function. */
};

/** @brief Structure used to enclose all elements required by the APIs to initialize a user device inside SEkey. */
typedef struct userdata_{
	std::string sn; /**< SEcube serial number */
	std::string uid; /**< user id */
	std::string uname; /**< username */
	uint32_t k1; /**< ID of symmetric key to encrypt user updates */
	uint32_t k2; /**< ID of symmetric key for key wrapping inside updates */
	uint32_t algo; /**< algorithm for encryption of updates */
	uint32_t klen; /**< length of keys */
	std::string query; /**< SQL query to be executed by the user */
	std::unique_ptr<uint8_t[]> k1_data; /**< content of k1 */
	std::unique_ptr<uint8_t[]> k2_data; /**< content of k2 */
	std::unique_ptr<uint8_t[]> wcard_key; /**< content of wildcard key (used to send an update to a former member of SEkey)*/
}userdata;

/** @brief Simlpe struct used to store the details about the SEkey user which is using the application.
 * ID and name are not really user when SEkey is in admin mode, they are used when SEkey is in
 * user mode. The device serial number instead is always used and is crucial for the correct execution
 * of the SEkey KMS. */
typedef struct userinfo_{
	std::string userid;
	std::string username;
	std::string device_sn;
} userinfo;

/** @brief The user class is used to model a user object. User's attributes are private but methods
 * are available to retrieve them. */
class se_user{
private:
	std::string id; /**< ID of the user */
	std::string name; /**< Name of the user. */
	std::string sn; /**< Serial number of the SEcube of the user. */
	std::string user_pin; /**< The PIN of the SEcube to login with low privilege (user mode, always disclosed to the user by the administrator). */
	std::string admin_pin; /**< The PIN of the SEcube to login with high privilege (admin mode, never disclosed to the user by the administrator). */
	std::string algorithm; /**< Algorithm used by the administrator to encrypt the SEkey updates for this user. */
	uint32_t k1; /**< ID of the key used to encrypt updates. */
	uint32_t k2; /**< ID of the key used to encrypt keys to be written inside the updates. */
	uint32_t init; /**< Flag that tells if the user SEcube is initialized or not. */
	int64_t update_cnt; /**< Ascending identifier of records written inside SEkey update files. */
	std::vector<std::string> groups; /**< List of groups to which the user belongs. */
public:
	se_user(): k1(0), k2(0), init(0), update_cnt(0){};
	se_user(std::string& _id, std::string& _name): id(_id), name(_name), k1(0), k2(0), init(0), update_cnt(0){};
	se_user(std::string& user_id, std::string& user_name, std::string& serialnumber, std::string& userpin, std::string& adminpin, uint32_t k1, uint32_t k2, uint32_t algo, uint32_t init_flag, int64_t cnt);
	void set_id(std::string& new_id); /**< Set the ID of the user. */
	std::string& get_id(); /**< Retrieve the ID of the user. */
	void set_name(std::string& new_name); /**< Set the name of the user. */
	void add_group(std::string& group); /**< Add a group to the list of the groups to which the user belongs. */
	void print_user_details(std::ofstream& sekey_log); /**< used for debugging purpose */
	std::string& get_sn(){return this->sn;};
};

/** @brief The policy class is used to model a security policy adopted by a group.
 * A policy for a group has 3 features: a maximum number of keys for the group,
 * a common algorithm for all keys belonging to the group, a default cryptoperiod
 * common to all keys belonging to the group. */
class group_policy{
private:
	uint32_t max_keys; /**< Maximum number of keys for the group. */
	uint32_t algorithm; /**< Algorithm used by all keys of this group. */
	uint32_t default_cryptoperiod; /**< Liveness of the keys of this group. */
	friend class se_group;
public:
	group_policy(uint32_t maxkeys, uint32_t algo, uint32_t cryptoperiod);
	group_policy();
	uint32_t get_max_keys();
	uint32_t get_algorithm();
	uint32_t get_default_cryptoperiod();
	void set_max_keys(uint32_t maxkeys);
	void set_default_cryptoperiod(uint32_t cryptoperiod);
	void set_algorithm(uint32_t algo);
	bool isvalid(); /**< A policy is valid if the algorithm is valid, the max number of keys is greater than 0 and the default cryptoperiod is greater than 0. */
};

/**
 * @brief The key class is used to model a key object. A key has several properties which
 * are kept private (id, name, owner, status, algorithm, length, generation time,
 * activation time, expiration time). Public methods are available to retrieve
 * the values of private properties.
*/
class se_key {
private:
	std::string id; /**< id of the key */
	std::string name; /**< label of the key */
	std::string owner; /**< the owner of a key is always a group */
	se_key_status status; /**< status (i.e. preactive) */
	se_key_type type; /**< the type of the key (i.e. symmetric data encryption) */
	uint32_t algorithm; /**< algorithm (must be the same specified by the policy of the group owner of the key) */
	uint32_t length; /**< given the algorithm, the length of the key (expressed in bits) is automatically determined */
	time_t generation; /**< time at which the key was created */
	time_t activation; /**< time at which the key was activated */
	time_t expiration; /**< time at which the key expired or was supposed to expire (deactivation is always less or equal to the expiration time) */
	time_t deactivation; /**< time at which the key was deactivated */
	time_t compromise; /**< time at which the key was compromised (0 if it is not compromised) */
	time_t destruction; /**< time at which the key was destroyed (0 if it is not destroyed) */
	time_t suspension; /**< last time at which the key was suspended (0 if it is not suspended) */
	time_t cryptoperiod; /**< time span during which a key can be used for encryption and decryption, if 0 the cryptoperiod of the owner is inherited */
public:
	se_key(): status(se_key_status::preactive), type(se_key_type::symmetric_data_encryption), algorithm(L1Algorithms::Algorithms::AES_HMACSHA256), length(0), generation(0),
		activation(0), expiration(0), deactivation(0), compromise(0), destruction(0), suspension(0), cryptoperiod(0){};
	se_key(std::string& key_id, uint32_t algo, uint32_t key_length, time_t act, time_t exp);
	se_key(std::string& key_id, std::string& key_name, std::string& key_owner, se_key_status key_status, uint32_t key_algo, uint32_t key_length, time_t gen, time_t act,
			time_t exp, time_t crypto, time_t deactivation, se_key_type key_type, time_t compr, time_t destr, time_t susp);
	se_key& operator= (const se_key& key);
	std::string& get_id();
	se_key_status get_status();
	bool safer(se_key& chosen); /**< Determines if the current key is safer than another key. */
	void print_key_details(std::ofstream& sekey_log); // used for debugging purpose
};

/** @brief Implement the concept of group inside SEkey.
 * Class members are kept private and suitable getter/setter methods are provided. */
class se_group{
private:
	std::string id; /**< ID of the group. */
	std::string name; /**< Name of the group. */
	uint32_t users_counter; /**< The number of users currently in the group. */
	uint32_t keys_counter; /**< The number of keys currently in the group. */
	group_policy policy; /**< The policy of the group. */
	std::vector<se_user> users_list; /**< Not used at the moment. */
	std::vector<se_key> keys_list; /**< Not used at the moment. */
public:
	se_group(std::string& groupid, std::string& groupname, group_policy gpolicy);
	se_group(): users_counter(0), keys_counter(0){};
	std::string& get_id();
	void set_id(std::string& new_id);
	std::string& get_name();
	void set_name(std::string& new_name);
	uint32_t get_users_counter();
	void set_users_counter(uint32_t cnt);
	uint32_t get_keys_counter();
	void set_keys_counter(uint32_t cnt);
	uint32_t get_keys_maxnumber();
	void set_keys_maxnumber(uint32_t max);
	uint32_t get_keys_algorithm();
	void set_keys_algorithm(uint32_t algo);
	uint32_t get_keys_cryptoperiod();
	void set_keys_cryptoperiod(uint32_t cryptoperiod);
	void print_group_details(std::ofstream& sekey_log); // used for debugging purpose
};

/** @brief Handy RAII wrapper for sqlite3_stmt which requires call to sqlite3_finalize to avoid resource leakage.
 *  See the SQLite documentation for more informations about sqlite3_stmt. */
class statement{
private:
	sqlite3_stmt *stmt; /**< Pointer to the statement as required by SQLite. */
public:
	/** The constructor will set the statement pointer to NULL, the statement will then be allocated by an explicit call to sqlite3_prepare(). */
	statement(){ this->stmt = nullptr; };
	/** Destructor involving sqlite3_finalize() to avoid memory leaks. */
	~statement(){ sqlite3_finalize(this->stmt); this->stmt = nullptr; };
	 /** Returns the pointer to the SQLite statement. */
	sqlite3_stmt *getstmt(){ return this->stmt; };
	 /** Returns the pointer to the pointer of the SQLite statement, implies finalization of previous pointer if not NULL (used for cyclig usage of same statement object). */
	sqlite3_stmt **getstmtref(){ sqlite3_finalize(this->stmt); return &(this->stmt); };
	/** Wrapper for sqlite3_finalize() SQLite API. */
	void finalize(){ sqlite3_finalize(this->stmt); };
};

/**
 * @defgroup sekey_apis
 *  @{
 */
	/**
	 * \name APIs of SEkey
	 * \details These are the APIs available to use SEkey.
	 */
	///@{
		int sekey_start(L0& l0, L1 *l1ptr);
		int sekey_stop();
		int sekey_admin_init(L1& l1, std::vector<std::string>& pin, std::string& userpin, std::string& adminpin);
		int sekey_init_user_SEcube(std::string& uid, std::string& userpin, std::string& adminpin, std::vector<std::string>& pin);
		int sekey_add_user(std::string& user_id, std::string& username);
		int sekey_delete_user(std::string& userID);
		int sekey_add_user_group(std::string& userID, std::string& groupID);
		int sekey_delete_user_group(std::string& user_id, std::string& group_id);
		int sekey_user_change_name(std::string& userID, std::string& newname);
		int sekey_user_get_info_all(std::vector<se_user> *users);
		int sekey_user_get_info(std::string& userid, se_user *user);
		int sekey_add_group(std::string& groupID, std::string& group_name, group_policy policy);
		int sekey_delete_group(std::string& groupID);
		int sekey_group_change_name(std::string& groupID, std::string& newname);
		int sekey_group_change_max_keys(std::string& groupID, uint32_t maxkeys);
		int sekey_group_change_default_cryptoperiod(std::string& groupID, uint32_t cryptoperiod);
		int sekey_group_get_info(std::string& groupID, se_group *group);
		int sekey_group_get_info_all(std::vector<se_group> *groups);
		int sekey_add_key(std::string& key_id, std::string& key_name, std::string& key_owner, uint32_t cryptoperiod, se_key_type keytype);
		int sekey_activate_key(std::string& key_id);
		int sekey_key_change_status(std::string& key_id, se_key_status status);
		int sekey_key_change_name(std::string& key_id, std::string& key_name);
		int sekey_key_get_info(std::string& key_id, se_key *key);
		int sekey_key_get_info_all(std::vector<se_key> *keys);
		int sekey_find_key_v1(std::string& chosen_key, std::string& source_user_id, std::string& dest_user_id, se_key_type keytype);
		int sekey_find_key_v2(std::string& chosen_key, std::string& source_user_id, std::string& group_id, se_key_type keytype);
		int sekey_find_key_v3(std::string& chosen_key, std::string& source_user_id, std::vector<std::string>& dest_user_id, se_key_type keytype);
		int sekey_readlog(std::string* sn, std::string& output);
	///@}
/** @}*/

/**
 * @defgroup sekey_other_functions
 *  @{
 */
	/**
	 * \name Other functions of SEkey
	 * \details These functions are used by SEkey for internal purposes but they can also be used by higher levels if needed. Always check the documentation and the source code before using them.
	 */
	///@{
time_t sekey_gettime();
int sekey_update_userdata();
int sekey_check_expired_keys();
int sekey_recovery_request(std::string& user_id, std::string& serial_number);
std::string epoch_to_localtime(time_t t);
std::string cryptoperiod_to_days(uint32_t n);
uint32_t stoul_wrap(std::string& s);
std::string statusmap(se_key_status s);
std::string keytypemap(se_key_type t);
std::string algomap(uint32_t algorithm);
uint32_t algolen(uint32_t algorithm);
	///@}
/** @}*/

/**
 * @defgroup sekey_internal_functions
 *  @{
 */
	/**
	 * \name Internal functions of SEkey
	 * \details These functions are used by SEkey for internal purposes so they should not be called expliticly.
	 */
	///@{
		void sekey_printlog(std::string& msg);
		int sekey_user_init(std::string& user_id, std::string& username, std::string& sn);
		int sekey_recovery();
		int sekey_write_recovery(std::string& user_id, std::string& serial_number);
		int reset_user_recovery(std::string& user_id, std::string& sn);
		void req_delete_user(std::string& user_id, std::string& uid, bool erase);
		void req_delete_user(std::string& user_id, uint32_t algo, uint32_t key_id, std::string& sn, bool erase, int mode);
		void req_delete_group(std::string& user_id, std::string& gid, bool erase);
		void req_delete_user_from_group(std::string& user_id, std::string& uid, std::string& group_id, bool erase);
		int usr_delete_user_from_group(char *buffer);
		int usr_delete_user(char *buffer);
		void send_key_update(std::string& user_id, uint32_t kid, uint32_t key_len, bool erase);
		int usr_delete_group(char *buffer);
		int usr_store_key(char *buffer);
		int send_user_init_update(std::string& user_id, std::string& query);
		void se3_flash_maintenance_routine();
		int usr_sql_exec(char *buffer, uint32_t bufsize);
		int execute_update(std::string& filepath);
		void send_sql_update(std::string& user_id, std::string& query, bool erase);
		int open_update_file(SEfile& updatefile, std::string& sn, bool overwrite, bool create, int mode);
		int check_key_transition_validity(se_key_status current_status, se_key_status new_status);
		int process_update_file();
		void delete_user_iterator(std::vector<std::string>& users, std::string& user_id, bool erase);
		void delete_group_iterator(std::vector<std::string>& users, std::string& group_id, bool erase);
		void sql_update_iterator(std::vector<std::string>& users, std::string& query, bool erase);
		void delete_user_from_group_iterator(std::vector<std::string>& users, std::string& user_id, std::string& group_id, bool erase);
		void key_update_iterator(std::vector<std::string>& users, uint32_t kid, uint32_t key_len, bool erase);
		int rollback_transaction();
		int commit_transaction();
		int fill_recovery(std::vector<std::string>& users);
		int sql_fill_vector(std::string *bind, std::string& query, std::vector<std::string> *container);
		uint32_t get_u32(sqlite3_stmt *stmt, int index);
		bool algovalid(uint32_t algorithm);
		int is_user_present(std::string& user_id);
		int is_group_present(std::string& group_id);
		int is_key_present(std::string& key_id);
		bool deletefile(SEfile *fileptr, std::string& filepath);
		int file_exists(std::string& filename);
		int generate_serial_number(char *sn);
		bool user_allowed();
		int algocmp(uint32_t algo1, uint32_t algo2);
		int sqlite3_expanded_sql_wrapper(sqlite3_stmt *stmt, std::string& s);
		std::string sqlite3_column_text_wrapper(sqlite3_stmt *stmt, int col);
		bool check_input(std::string& in, uint8_t sel);
	///@}
/** @}*/

#endif
